#ifdef HAVE_LIBCAP
#include <stdlib.h>
#include <error.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <linux/types.h>
#include <sys/prctl.h>

#include "modcap.h"

#define FAIL_EXIT(fstr, args...) error (EXIT_FAILURE, errno, fstr, ##args)

int modify_capability_common(cap_value_t cap, cap_flag_value_t on)
{
	cap_t cap_p = cap_get_proc();
	cap_flag_value_t cap_ok;
	int rc = -1;

	if (!cap_p) {
		FAIL_EXIT("cap_get_proc");
	}

	cap_ok = CAP_CLEAR;
	if (cap_get_flag(cap_p, cap, CAP_PERMITTED, &cap_ok) < 0) {
		FAIL_EXIT("cap_get_flag");
	}

	if (cap_ok == CAP_CLEAR) {
		rc = on ? -1 : 0;
		goto out;
	}

	if (cap_set_flag(cap_p, CAP_EFFECTIVE, 1, &cap, on) < 0) {
		FAIL_EXIT("cap_set_flag");
	}

	if (cap_set_proc(cap_p) < 0) {
		FAIL_EXIT("cap_set_proc");
	}

	rc = 0;
out:
	if (cap_p)
		cap_free(cap_p);
	return rc;
}

void limit_capabilities_common(int perm_cap_raw, int perm_cap_admin)
{
	cap_t cap_cur_p;
	cap_t cap_p;
	cap_flag_value_t cap_ok;
	const cap_value_t caps[] = {CAP_SETUID, CAP_SETGID};
	const cap_value_t cap_raw = CAP_NET_RAW;
	const cap_value_t cap_admin = CAP_NET_ADMIN;
	uid_t uid = getuid();

	cap_p = cap_init();
	if (!cap_p) {
		FAIL_EXIT("cap_init");
	}

	if (!uid) {
		const char *user = "iputils";
		const char *dir = "/var/resolv";
		struct passwd *pw = getpwnam(user);
		gid_t gid;

		if (!pw)
			FAIL_EXIT("lookup of user \"%s\" failed", user);

		gid = pw->pw_gid;
		uid = pw->pw_uid;
		endpwent();

		if (!uid)
			FAIL_EXIT("user \"%s\" shouldn't be root", user);

		if (initgroups(user, gid) < 0)
			FAIL_EXIT("initgroups");

		if (chroot(dir) < 0)
			FAIL_EXIT("chroot to \"%s\" failed", dir);

		if (chdir("/") < 0)
			FAIL_EXIT("chdir to \"/\" failed");

		if (cap_set_flag(cap_p, CAP_PERMITTED, 2, caps, CAP_SET) < 0 ||
				cap_set_flag(cap_p, CAP_EFFECTIVE, 2, caps, CAP_SET) < 0) {
			FAIL_EXIT("cap_set_flag");
		}

		if (setregid (gid, gid) < 0) {
			FAIL_EXIT("setregid");
		}
	}

	cap_cur_p = cap_get_proc();
	if (!cap_cur_p) {
		FAIL_EXIT("cap_get_proc");
	}

	if (perm_cap_admin) {
		cap_ok = CAP_CLEAR;
		if (cap_get_flag(cap_cur_p, CAP_NET_ADMIN, CAP_PERMITTED, &cap_ok) < 0)
			FAIL_EXIT("cap_get_flag");

		if (cap_ok != CAP_CLEAR) {
			if (cap_set_flag(cap_p, CAP_PERMITTED, 1, &cap_admin, CAP_SET) < 0)
				FAIL_EXIT("cap_set_flag");
		}
	}

	if (perm_cap_raw) {
		cap_ok = CAP_CLEAR;
		if (cap_get_flag(cap_cur_p, CAP_NET_RAW, CAP_PERMITTED, &cap_ok) < 0)
			FAIL_EXIT("cap_get_flag");

		if (cap_ok != CAP_CLEAR) {
			if (cap_set_flag(cap_p, CAP_PERMITTED, 1, &cap_raw, CAP_SET) < 0)
				FAIL_EXIT("cap_set_flag");
		}
	}

	cap_free(cap_cur_p);

	if (cap_set_proc(cap_p) < 0) {
		FAIL_EXIT("cap_set_proc");
	}

	if (prctl(PR_SET_KEEPCAPS, 1) < 0) {
		FAIL_EXIT("prctl(PR_SET_KEEPCAPS, 1) failed");
	}

	if (setreuid (uid, uid) < 0) {
		FAIL_EXIT("setreuid");
	}

	if (prctl(PR_SET_KEEPCAPS, 0) < 0) {
		FAIL_EXIT("prctl");
	}

	if (cap_set_flag(cap_p, CAP_EFFECTIVE, 2, caps, CAP_CLEAR) < 0 ||
			cap_set_flag(cap_p, CAP_PERMITTED, 2, caps, CAP_CLEAR) < 0) {
		FAIL_EXIT("cap_set_flag");
	}

	if (cap_set_proc(cap_p) < 0) {
		FAIL_EXIT("cap_set_proc");
	}

	cap_free(cap_p);
}

void drop_capabilities_common(void)
{
	cap_t cap = cap_init();
	if (cap_set_proc(cap) < 0) {
		FAIL_EXIT("cap_set_proc");
	}
	cap_free(cap);
}
#endif /* HAVE_LIBCAP */
